@@ -62,48 +62,69 @@ List project_to_curve(NumericMatrix x, NumericMatrix s, double stretch = 2) {
     stop("Argument 'stretch' should be larger than or equal to 0");
   }
 
-  // precompute distances between successive points in the curve
-  // and the length of each segment
-  int num_segments = s.nrow() - 1;
-  NumericMatrix diff(num_segments, s.ncol());
-  NumericVector length(num_segments);
+  int nseg = s.nrow() - 1;
+  int npts = x.nrow();
+  int ncols = x.ncol();
 
-  for (int i = 0; i < num_segments; ++i) {
-    diff(i, _) = s(i + 1, _) - s(i, _);
-    length[i] = sum(pow(diff(i, _), 2.0));
+  if (s.ncol() != ncols) {
+    stop("'x' and 's' must have an equal number of columns");
   }
 
-  // OUTPUT DATA STRUCTURES
-  // projections of x onto s
-  NumericMatrix new_s(x.nrow(), x.ncol());
-  rownames(new_s) = rownames(x);
-  colnames(new_s) = colnames(x);
+  // precompute distances between successive points in the curve
+  // and the length of each segment
+  NumericMatrix diff = no_init(nseg, ncols);
+  NumericVector length = no_init(nseg);
 
-  // distance from start of the curve
-  NumericVector lambda(x.nrow());
+  for (int i = 0; i < nseg; ++i) {
+    // OPTIMISATION: compute length manually
+    //   diff(i, _) = s(i + 1, _) - s(i, _);
+    //   length[i] = sum(pow(diff(i, _), 2));
+    double l = 0;
+    for (int k = 0; k < ncols; ++k) {
+      double value = s(i + 1, k) - s(i, k);
+      diff(i, k) = value;
+      l += value * value;
+    }
+    length[i] = l;
+    // END OPTIMISATION
+  }
 
-  // distances between x and new_s
-  NumericVector dist_ind(x.nrow());
-  dist_ind.attr("names") = rownames(x);
+  // allocate output data structures
+  NumericMatrix new_s = no_init(npts, ncols);     // projections of x onto s
+  NumericVector lambda = no_init(npts);           // distance from start of the curve
+  NumericVector dist_ind = no_init(npts);         // distances between x and new_s
+
+  // pre-allocate intermediate vectors
+  NumericVector n_test = no_init(ncols);
+  NumericVector n = no_init(ncols);
+  NumericVector p = no_init(ncols);
 
   // iterate over points in x
-  for (int i = 0; i < x.nrow(); ++i) {
-    NumericVector p = x(i, _);
-
+  for (int i = 0; i < npts; ++i) {
     // store information on the closest segment
-    int bestj = -1;
-    double bestt = -1;
-    NumericVector n(x.ncol());
+    double bestlam = -1;
     double bestdi = R_PosInf;
 
+    // copy current point to p
+    for (int k = 0; k < ncols; ++k) {
+      p[k] = x(i, k);
+    }
+
     // iterate over the segments
-    for (int j = 0; j < num_segments; ++j) {
+    for (int j = 0; j < nseg; ++j) {
 
       // project p orthogonally onto the segment
-      NumericVector diff1 = diff(j, _);
-      NumericVector diff2 = p - s(j, _);
+      // OPTIMISATION: do not allocate diff1 and diff2; compute t manually
+      //   NumericVector diff1 = s(j + 1, _) - s(j, _);
+      //   NumericVector diff2 = p - s(j, _);
+      //   double t = sum(diff1 * diff2) / length(j);
+      double t = 0;
+      for (int k = 0; k < ncols; ++k) {
+        t += diff(j, k) * (p[k] - s(j, k));
+      }
+      t /= length(j);
+      // END OPTIMISATION
 
-      double t = sum(diff1 * diff2) / length(j);
       if (t < 0) {
         t = 0.0;
       }
@@ -112,24 +133,34 @@ List project_to_curve(NumericMatrix x, NumericMatrix s, double stretch = 2) {
       }
 
       // calculate position of projection and the distance
-      NumericVector n_test = s(j, _) + t * diff(j, _);
-
-      // calcualte distance of projection and original point
-      double di = sum(pow(n_test - p, 2.0));
+      // OPTIMISATION: compute di and n_test manually
+      //   NumericVector n_test = s(j, _) + t * diff(j, _);
+      //   double di = sum(pow(n_test - p, 2.0));
+      double di = 0;
+      for (int k = 0; k < ncols; ++k) {
+        double value = s(j, k) + t * diff(j, k);
+        n_test(k) = value;
+        di += (value - p[k]) * (value - p[k]);
+      }
+      // END OPTIMISATION
 
       // if this is better than what was found earlier, store it
       if (di < bestdi) {
         bestdi = di;
-        n = n_test;
-        bestj = j;
-        bestt = t;
+        bestlam = j + .1 + .9 * t;
+        for (int k = 0; k < ncols; ++k) {
+          n[k] = n_test[k];
+        }
+
       }
     }
 
     // save the best projection to the output data structures
-    new_s(i, _) = n;
-    lambda[i] = bestj + .1 + .9 * bestt;
+    lambda[i] = bestlam;
     dist_ind[i] = bestdi;
+    for (int k = 0; k < ncols; ++k) {
+      new_s(i, k) = n[k];
+    }
   }
 
   // get ordering from old lambda
@@ -138,23 +169,39 @@ List project_to_curve(NumericMatrix x, NumericMatrix s, double stretch = 2) {
   // calculate total dist
   double dist = sum(dist_ind);
 
-  // calculate lambda for new_s
-  NumericVector new_lambda(new_ord.length());
-  new_lambda.attr("names") = rownames(x);
+  // recalculate lambda for new_s
+  lambda[new_ord[0]] = 0;
 
   for (int i = 1; i < new_ord.length(); ++i) {
     int o1 = new_ord[i];
     int o0 = new_ord[i - 1];
-    NumericVector p1 = new_s(o1, _);
-    NumericVector p0 = new_s(o0, _);
-    new_lambda[o1] = new_lambda[o0] + sqrt(sum(pow(p1 - p0, 2.0)));
+
+    // OPTIMISATION: compute lambda[o1] manually
+    //   NumericVector p1 = new_s(o1, _);
+    //   NumericVector p0 = new_s(o0, _);
+    //   lambda[o1] = lambda[o0] + sqrt(sum(pow(p1 - p0, 2.0)));
+    double o0o1 = 0;
+    for (int k = 0; k < ncols; ++k) {
+      double val = new_s(o1, k) - new_s(o0, k);
+      o0o1 += val * val;
+    }
+    lambda[o1] = lambda[o0] + sqrt(o0o1);
+    // END OPTIMISATION
   }
 
+  // make sure all dimnames are correct
+  new_s.attr("dimnames") = List::create(
+    rownames(x),
+    colnames(x)
+  );
+  dist_ind.attr("names") = rownames(x);
+  lambda.attr("names") = rownames(x);
+
   // return output
   List ret;
   ret["s"] = new_s;
   ret["ord"] = new_ord + 1;
-  ret["lambda"] = new_lambda;
+  ret["lambda"] = lambda;
   ret["dist_ind"] = dist_ind;
   ret["dist"] = dist;
 
